# coding: utf-8
import os
from datetime import datetime

import pytest

from pycrunch.shoji import as_entity, wait_progress, TaskError

from scrunch.datasets import Project
from scrunch.helpers import shoji_entity_wrapper
from scrunch.scripts import ScriptExecutionError
from scrunch.mutable_dataset import get_mutable_dataset
from fixtures import BaseIntegrationTestCase


@pytest.mark.skipif(os.environ.get("LOCAL_INTEGRATION") is None, reason="Do not run this test during CI/CD")
class TestSystemScripts(BaseIntegrationTestCase):
    def new_project(self, name):
        res = self.site.projects.create(shoji_entity_wrapper({
            "name": name + datetime.now().strftime("%Y%m%d%H%M%S")
        })).refresh()
        return Project(res)

    def test_define_view_strict_subvariable_syntax(self):
        project = self.new_project("test_view_strict_subvariable")
        ds = self.site.datasets.create(as_entity({"name": "test_dataset_script"})).refresh()
        categories = [
            {"id": 2, "name": "Home"},
            {"id": 3, "name": "Work"},
            {"id": -1, "name": "No Data", "missing": True},
        ]
        subvariables = [
            {"alias": "cat", "name": "Cat"},
            {"alias": "dog", "name": "Dog"},
            {"alias": "bird", "name": "Bird"},
        ]

        ds.variables.create(
            as_entity(
                dict(
                    alias="pets",
                    name="Pets",
                    type="categorical_array",
                    categories=categories,
                    subvariables=subvariables,
                    values=[[2, 3, 3], [3, 3, 2], [2, -1, 3], [3, 2, -1]],
                )
            )
        )
        ds.variables.create(
            as_entity(
                dict(
                    alias="pets_2",
                    name="Pets 2",
                    type="categorical_array",
                    categories=categories,
                    subvariables=subvariables,
                    values=[[2, 3, 3], [3, 3, 2], [2, -1, 3], [3, 2, -1]],
                )
            )
        )
        script_body = """
                        DEFINE VIEW FROM DATASET_ID(`{}`)
                        VARIABLES pets, pets_2
                        NAME "My view";
                        """.format(ds.body.id)

        scrunch_dataset = get_mutable_dataset(ds.body.id, self.site)
        project.move_here([scrunch_dataset])
        resp = project.execute(script_body, strict_subvariable_syntax=True)
        wait_progress(resp, self.site.session)
        view = scrunch_dataset.views.get_by_name("My view")
        assert view.project.name == project.name

    def test_define_view_strict_subvariable_syntax_error(self):
        project = self.new_project("test_view_strict_subvariable_false")
        ds = self.site.datasets.create(as_entity({"name": "test_dataset_script_false"})).refresh()
        categories = [
            {"id": 2, "name": "Home"},
            {"id": 3, "name": "Work"},
            {"id": -1, "name": "No Data", "missing": True},
        ]
        subvariables = [
            {"alias": "cat", "name": "Cat"},
            {"alias": "dog", "name": "Dog"},
            {"alias": "bird", "name": "Bird"},
        ]

        ds.variables.create(
            as_entity(
                dict(
                    alias="pets",
                    name="Pets",
                    type="categorical_array",
                    categories=categories,
                    subvariables=subvariables,
                    values=[[2, 3, 3], [3, 3, 2], [2, -1, 3], [3, 2, -1]],
                )
            )
        )
        ds.variables.create(
            as_entity(
                dict(
                    alias="pets_2",
                    name="Pets 2",
                    type="categorical_array",
                    categories=categories,
                    subvariables=subvariables,
                    values=[[2, 3, 3], [3, 3, 2], [2, -1, 3], [3, 2, -1]],
                )
            )
        )
        script_body = """
                                DEFINE VIEW FROM DATASET_ID(`{}`)
                                VARIABLES pets, pets_2
                                NAME "My view";
                                """.format(ds.body.id)

        try:
            scrunch_dataset = get_mutable_dataset(ds.body.id, self.site)
            project.move_here([scrunch_dataset])
            resp = project.execute(script_body)
            with pytest.raises(TaskError) as err: 
                wait_progress(resp, self.site.session)
            err_value = err.value[0]
            err_value["type"] == "script:validation"
            err_value["description"] == "Errors processing the script"
            err_value["resolutions"][0]["message"] == "The following subvariables: bird, cat, dog exist in multiple arrays: pets, pets_2"
        finally:
            ds.delete()
            project.delete()


class TestDatasetScripts(BaseIntegrationTestCase):
    def _create_ds(self):
        ds = self.site.datasets.create(as_entity({"name": "test_script"})).refresh()
        variable = ds.variables.create(
            as_entity(
                {
                    "name": "pk",
                    "alias": "pk",
                    "type": "numeric",
                }
            )
        )
        return ds, variable

    def test_execute(self):
        ds, variable = self._create_ds()
        try:
            scrunch_dataset = get_mutable_dataset(ds.body.id, self.site, editor=True)
            script = """
            RENAME pk TO varA;

            CHANGE TITLE IN varA WITH "Variable A";
            """
            scrunch_dataset.scripts.execute(script)
            variable.refresh()
            assert variable.body["alias"] == "varA"
            assert variable.body["name"] == "Variable A"
        finally:
            ds.delete()

    def test_handle_error(self):
        ds, variable = self._create_ds()
        scrunch_dataset = get_mutable_dataset(ds.body.id, self.site)
        script = """BAD-RENAME pk TO varA;"""  # Bad syntax script
        with pytest.raises(ScriptExecutionError) as err:
            scrunch_dataset.scripts.execute(script)

        variable.refresh()
        assert variable.body["alias"] == "pk"  # Unchanged
        resolutions = err.value.resolutions[0]
        assert resolutions["column"] == 1
        assert resolutions["line"] == 1
        assert resolutions["command"] == 1
        assert resolutions["message"] == "Invalid command: BAD"

        # Script big enough to trigger async validation
        async_script = ["""BAD-RENAME pk TO varA;"""] * 50000
        async_script = "\n".join(async_script)
        assert len(async_script) > 2 ** 20  # This is the threshold for async
        with pytest.raises(ScriptExecutionError) as err:
            scrunch_dataset.scripts.execute(async_script)
        assert len(err.value.resolutions) == 50000  # All lines raised error
        resolutions = err.value.resolutions[0]
        assert resolutions["column"] == 1
        assert resolutions["line"] == 1
        assert resolutions["command"] == 1
        assert resolutions["message"] == "Invalid command: BAD"
        ds.delete()

    def test_revert_script(self):
        ds, variable = self._create_ds()
        scrunch_dataset = get_mutable_dataset(ds.body.id, self.site)
        s1 = "RENAME pk TO varA;"
        s2 = 'CHANGE TITLE IN varA WITH "Variable A";'

        scrunch_dataset.scripts.execute(s1)
        scrunch_dataset.scripts.execute(s2)

        scrunch_dataset.scripts.revert_to(script_number=0)  # Reverted
        variable.refresh()
        assert variable.body["name"] == "pk"
        ds.delete()

    @pytest.mark.skip(reason="Collapse is 504ing in the server.")
    def test_fetch_all_and_collapse(self):
        ds, variable = self._create_ds()
        try:
            scrunch_dataset = get_mutable_dataset(ds.body.id, self.site)
            s1 = "RENAME pk TO varA;"
            s2 = 'CHANGE TITLE IN varA WITH "Variable A";'

            scrunch_dataset.scripts.execute(s1)
            scrunch_dataset.scripts.execute(s2)

            r = scrunch_dataset.scripts.all()
            assert len(r) == 2
            assert r[0].body["body"] == s1
            assert r[1].body["body"] == s2

            scrunch_dataset.scripts.collapse()

            r = scrunch_dataset.scripts.all()
            assert len(r) == 1
        finally: 
            ds.delete()



try:
    # py <= 3.7
    from importlib_metadata import version
except ModuleNotFoundError:
    # py >= 3.8
    from importlib.metadata import version


try:
    __version__ = version('scrunch')
except Exception:
    # Package is installed from source. It's at least this version
    __version__ = '0.18.5-uninstalled'



from .session import connect
from .datasets import (
    get_user, get_project, get_dataset, get_team, create_team)
from .streaming_dataset import get_streaming_dataset
from .mutable_dataset import get_mutable_dataset, create_dataset
from .version import __version__


__all__ = [
    'connect', 'get_user', 'get_project', 'get_dataset', 'get_team',
    'get_streaming_dataset', 'get_mutable_dataset', 'create_team',
    'create_dataset', '__version__'
]



# coding: utf-8

from scrunch.helpers import shoji_entity_wrapper


class DatasetViews:
    def __init__(self, dataset_resource):
        self.dataset_resource = dataset_resource

    def create(self, name, columns):
        """
        Creates a new View on this dataset

        :param name: Name of the view
        :param columns: Use None to copy all views, otherwise a list of aliases
        :return: Mutable Dataset for the new view
        """
        view_args = {
            "name": name,
            "view_of": self.dataset_resource.self,
            "owner": self.dataset_resource.body["owner"]
        }
        if columns is not None:
            # Columns is a list of aliases, convert to URLs
            alias_2_url = self.dataset_resource.variables.by("alias")
            columns_url = [alias_2_url[a].entity_url for a in columns]
            view_args["view_cols"] = columns_url

        project = self.dataset_resource.session.root.datasets
        view_res = project.create(shoji_entity_wrapper(view_args))
        view_res.refresh()

        from scrunch.mutable_dataset import MutableDataset
        return MutableDataset(view_res)

    def get_by_name(self, view_name):
        """
        Returns a view by name as a MutableDataset
        """
        from scrunch.mutable_dataset import MutableDataset
        views_by_name = self.dataset_resource.follow("views").by("name")
        view_res = views_by_name[view_name].entity
        return MutableDataset(view_res)

    def all_names(self):
        """
        Returns a list of the names of available views on this dataset.
        """
        views_index = self.dataset_resource.follow("views").index
        return {t["name"] for t in views_index.values()}



# coding: utf-8

import json
from requests import Response
from scrunch.accounts import Account
from scrunch.datasets import Project
from pycrunch.shoji import Entity, Catalog, View

from .mock_session import MockSession


class TestAccount:
    def make_session(self):
        root_url = "http://host/api/"
        account_url = "http://host/api/account/"
        projects_url = "http://host/api/account/projects/"
        execute_url = "http://host/api/account/run/"

        session = MockSession(site_url=root_url)
        root_resource = Catalog(session, **{
            "element": "shoji:catalog",
            "self": root_url,
            "index": {},
            "catalogs": {},
            "views": {"account": account_url}
        })
        act_resource = Entity(session, **{
            "element": "shoji:entity",
            "self": account_url,
            "body": {
                "name": "test account",
                "id": "00001",
            },
            "catalogs": {
                "projects": projects_url
            },
            "views": {
                "execute": execute_url
            }
        })
        execute_resource = View(session, **{
            "element": "shoji:view",
            "self": execute_url,
            "value": {}
        })
        session.add_fixture(account_url, act_resource)
        session.add_fixture(root_url, root_resource)
        session.add_fixture(execute_url, execute_resource)
        return session

    def test_get_current(self):
        api_root = self.make_session().root
        current_act = Account.current_account(api_root)
        assert current_act.name == "test account"
        assert current_act.id == "00001"
        assert isinstance(current_act, Account)

    def test_execute(self):
        session = self.make_session()

        response = Response()
        response.status_code = 204

        session.add_post_response(response)
        current_act = Account.current_account(session.root)

        current_act.execute("NOOP;")
        post_request = session.requests[-1]
        assert json.loads(post_request.body) == {
            "element": "shoji:view",
            "value": "NOOP;"
        }

    def test_execute_script_with_syntax_subvariable_flag(self):
        session = self.make_session()

        response = Response()
        response.status_code = 204

        session.add_post_response(response)
        current_act = Account.current_account(session.root)

        current_act.execute("NOOP;", strict_subvariable_syntax=True)
        post_request = session.requests[-1]
        assert json.loads(post_request.body) == {
            "element": "shoji:view",
            "value": "NOOP;"
        }
        assert "?strict_subvariable_syntax=true" in post_request.url

    def test_projects(self):
        session = self.make_session()
        projects_url = "http://host/api/account/projects/"
        project_url = "http://host/api/projects/abc/"

        project_resource = Entity(session, **{
            "element": "shoji:entity",
            "self": project_url,
            "body": {
                "name": "my project",
                "id": "abc"
            }
        })

        act_projects_res = Catalog(session, **{
            "element": "shoji:catalog",
            "self": projects_url,
            "index": {
                project_url: {
                    "name": project_resource["body"]["name"],
                    "id": project_resource["body"]["id"],
                }
            }
        })
        session.add_fixture(projects_url, act_projects_res)
        session.add_fixture(project_url, project_resource)
        current_act = Account.current_account(session.root)

        project = current_act.projects.by_name("my project")
        assert project.url == project_url
        assert isinstance(project, Project)

        project = current_act.projects.by_id("abc")
        assert project.url == project_url
        assert isinstance(project, Project)



# coding: utf-8

"""
Tests that the variable Folders API is properly supported

"""

from pycrunch.shoji import Catalog

from fixtures import NEWS_DATASET, BaseIntegrationTestCase
from scrunch import get_dataset
from scrunch.datasets import Variable
from scrunch.exceptions import InvalidPathError
from scrunch.folders import Folder


def setup_folders(ds):
    sess = ds.session
    ds.settings.edit(variable_folders=True)
    ds.variables.create({
        'element': 'shoji:entity',
        'body': {'name': 'testvar1', 'type': 'numeric'}
    })
    ds.variables.create({
        'element': 'shoji:entity',
        'body': {'name': 'testvar2', 'type': 'numeric'}
    })
    ds.variables.create({
        'element': 'shoji:entity',
        'body': {'name': 'testvar3', 'type': 'numeric'}
    })
    ds.variables.create({
        'element': 'shoji:entity',
        'body': {'name': 'testvar4', 'type': 'numeric'}
    })
    ds.variables.create({
        'element': 'shoji:entity',
        'body': {'name': 'testvar5', 'type': 'numeric'}
    })
    ds.variables.create({
        'element': 'shoji:entity',
        'body': {'name': 'testvar6', 'type': 'numeric'}
    })
    ds.refresh()
    sf1 = Catalog(sess, body={
        'name': 'Subfolder 1'
    })
    sf1 = ds.folders.public.create(sf1)
    sfa = Catalog(sess, body={
        'name': 'Subfolder A'
    })
    sfa = sf1.create(sfa)
    sf2 = Catalog(sess, body={
            'name': 'Subfolder 2'
        })
    sf2 = ds.folders.public.create(sf2)
    variables = ds.variables.by('alias')
    sf1.patch({'index': {
        variables['age'].entity_url: {}
    }})
    sfa.patch({'index': {
        variables['gender'].entity_url: {}
    }})
    sf2.patch({'index': {
        variables['socialmedia'].entity_url: {}
    }})


class TestFolders(BaseIntegrationTestCase):
    def setUp(self):
        super(TestFolders, self).setUp()
        self._ds = self.site.datasets.create({
            'element': 'shoji:entity',
            'body': {
                'name': 'test_folders',
                'table': {
                    'element': 'crunch:table',
                    'metadata': NEWS_DATASET
                },
            }
        }).refresh()
        ds = self._ds
        setup_folders(ds)
        self.ds = get_dataset(ds.body.id)

    def test_get_folders(self):
        ds = self.ds
        public = ds.folders.get('|')
        sf1 = ds.folders.get('| Subfolder 1')
        sfa = ds.folders.get('| Subfolder 1 | Subfolder A')

        # Equivalent ways of fetching Subfolder A
        sfa2 = public.get('Subfolder 1 | Subfolder A')
        sfa3 = sf1.get('Subfolder A')
        self.assertEqual(sfa.url, sfa2.url)
        self.assertEqual(sfa.url, sfa3.url)

        # Fetching a variable by path
        variable = ds.folders.get('| Subfolder 1 | Subfolder A | Gender')

        self.assertTrue(isinstance(sf1, Folder))
        self.assertTrue(isinstance(sfa, Folder))
        self.assertTrue(isinstance(variable, Variable))

        self.assertEqual(sf1.name, 'Subfolder 1')
        self.assertEqual(sfa.name, 'Subfolder A')
        self.assertEqual(sf1.parent, public)
        self.assertEqual(sfa.parent.name, sf1.name)
        self.assertEqual(sfa.parent.path, sf1.path)
        self.assertEqual(sfa.path, '| Subfolder 1 | Subfolder A')
        self.assertEqual(variable.alias, 'gender')
        self.assertEqual(variable.type, 'categorical')

        bad_path = '| bad folder'
        with self.assertRaises(InvalidPathError) as err:
            ds.folders.get(bad_path)
        self.assertEqual(str(err.exception), "Invalid path: %s" % bad_path)

    def test_create_folder(self):
        ds = self.ds
        public = ds.folders.get('|')
        mit = public.create_folder('Made in test')
        self.assertEqual(mit.path, "| Made in test")
        mit2 = public.get(mit.name)
        self.assertEqual(mit2.url, mit.url)
        nested = mit.create_folder('nested level')
        self.assertEqual(mit.get_child(nested.name).url, nested.url)

    def test_create_folder_with_variables(self):
        ds = self.ds
        public = ds.folders.public
        sf = public.create_folder("with children", alias=['testvar1', 'testvar2'])
        self.assertEqual([Variable, Variable], [type(c) for c in sf.children])
        self.assertEqual(['testvar1', 'testvar2'],
            [c.alias for c in sf.children])

    def test_reorder_folder(self):
        ds = self.ds
        public = ds.folders.get('|')
        folder = public.create_folder('ToReorder')
        sf1 = folder.create_folder('1')
        sf2 = folder.create_folder('2')
        sf3 = folder.create_folder('3')
        var = ds['testvar1']
        folder.move_here([var])
        children = folder.children
        self.assertEqual([c.url for c in children],
            [c.url for c in [sf1, sf2, sf3, var]])
        self.assertEqual(list(map(type, children)),
                         [Folder, Folder, Folder, Variable])

        # Reorder placing sf1 at the end
        folder.reorder([sf2, var, sf3, sf1])
        children = folder.children
        self.assertEqual([c.url for c in children],
            [c.url for c in [sf2, var, sf3, sf1]])

        # Test reorder providing children names only
        folder.reorder(var.alias, sf1.name, sf2.name, sf3.name)
        self.assertEqual([c.url for c in folder.children],
            [c.url for c in [var, sf1, sf2, sf3]])

    def test_move_to_folder(self):
        ds = self.ds
        sf = ds.folders.public.create_folder("target")
        ds['testvar1'].move_to_folder(sf.path)
        self.assertEqual(['testvar1'], [c.alias for c in sf.children])

    def test_move_between_folders(self):
        public = self.ds.folders.public
        target1 = public.create_folder("t1")
        target2 = public.create_folder("t2")
        self.assertEqual(target1.children, [])
        self.assertEqual(target2.children, [])
        var1 = self.ds['testvar2']
        var2 = self.ds['testvar3']
        target1.move_here([var2, var1])
        self.assertEqual([c.url for c in target1.children],
            [var2.url, var1.url])

        nested = target1.create_folder('nested')
        self.assertEqual([c.url for c in target1.children],
            [var2.url, var1.url, nested.url])

        # Move var2 to t2
        target2.move_here(var2)
        # Now, t1 does not have var2, it is under t2
        self.assertEqual([c.url for c in target1.children],
            [var1.url, nested.url])
        self.assertEqual([c.url for c in target2.children],
            [var2.url])

        # Move nested and var1 to t2
        target2.move_here([nested, var1])

        # Now, they are all in t2, t1 is empty
        self.assertEqual([c.url for c in target2.children],
            [var2.url, nested.url, var1.url])
        self.assertEqual([c.url for c in target1.children], [])

        # Move var1 to nested
        nested.move_here(var1)

        # Now, t2 doesn't have var1, but nested does
        self.assertEqual([c.url for c in target2.children],
            [var2.url, nested.url])
        self.assertEqual([c.url for c in nested.children],
            [var1.url])

        # Move nested to t1
        target1.move_here(nested)

        # Now t2 has only var2 and t1 has nested in it
        self.assertEqual([c.url for c in target2.children],
            [var2.url])
        self.assertEqual([c.url for c in target1.children],
            [nested.url])

    def test_move_here_position(self):
        public = self.ds.folders.public
        sf = public.create_folder("here")
        sf.move_here(self.ds['testvar1'])
        sf.move_here(self.ds['testvar2'], before='testvar1')
        self.assertEqual([c.name for c in sf.children], ['testvar2', 'testvar1'])
        sf.move_here(self.ds['testvar3'], after='testvar2')
        self.assertEqual([c.name for c in sf.children], ['testvar2', 'testvar3', 'testvar1'])
        sf.move_here(self.ds['testvar4'], position=1)
        self.assertEqual([c.name for c in sf.children], ['testvar2', 'testvar4', 'testvar3', 'testvar1'])

    def test_move_by_alias(self):
        public = self.ds.folders.public
        target = public.create_folder("test_move_by_alias")
        target.move_here('testvar2')
        self.assertEqual([c.name for c in target.children], ["testvar2"])

    def test_make_folder_in_position(self):
        public = self.ds.folders.public.create_folder('testhere')
        public.create_folder("p1")
        public.create_folder("p2")
        public.create_folder("p3")

        public.create_folder("A", before="p2")
        self.assertEqual([c.name for c in public.children],
            ["p1", "A", "p2", "p3"])

        public.create_folder("B", after="p2")
        self.assertEqual([c.name for c in public.children],
            ["p1", "A", "p2", "B", "p3"])

        public.create_folder("C", position=3)
        self.assertEqual([c.name for c in public.children],
            ["p1", "A", "p2", "C", "B", "p3"])

    def test_hide_variables(self):
        hidden_folder = self.ds.folders.hidden
        public = self.ds.folders.public
        var1 = self.ds['testvar4']
        var2 = self.ds['testvar5']
        public.move_here(var1, var2)
        hidden_folder.move_here(var2)

        var1_id = var1.resource.body.id
        var2_id = var2.resource.body.id
        self.assertTrue(var2_id in self._ds.folders.hidden.by('id'))
        self.assertTrue(var2_id not in self._ds.folders.public.by('id'))
        self.assertTrue(var1_id in self._ds.folders.public.by('id'))
        self.assertTrue(var1_id not in self._ds.folders.hidden.by('id'))

    def test_rename(self):
        public = self.ds.folders.public
        sf = public.create_folder('rename me')
        sf.rename("renamed")
        self.assertTrue('renamed' in [c.name for c in public.children])
        self.assertEqual(sf.name, 'renamed')

    def test_delete_folder(self):
        public = self.ds.folders.public
        var = self.ds['testvar6']
        sf = public.create_folder('delete me')
        sf.move_here(var)
        self.assertTrue(var.url in self._ds.variables.index)
        sf.delete()
        # Folder isn't in public anymore
        self.assertFalse(sf.url in [c.url for c in public.children])

    def test_dict_protocol(self):
        public = self.ds.folders.public
        sf1 = public.create_folder("level1")
        sf1a = public.create_folder("level1a")
        sf2 = sf1.create_folder("level2")

        # Access by path []
        self.assertEqual(public[sf2.path].url, sf2.url)

        # Access by name
        self.assertEqual(public[sf1.name].url, sf1.url)
        self.assertEqual(sf1[sf2.name].url, sf2.url)

        # __iter__ method
        self.assertEqual([c.url for c in public], [c.url for c in public.children])

        # more dict methods
        self.assertEqual(set(dict(public).keys()), set(public.keys()))
        self.assertEqual({c.url for c in dict(public).values()},
            {c.url for c in public.children})
        # The right key corresponds to the correct value
        self.assertEqual(dict(public)[sf1.name].url, sf1.url)

    def test_dict_protocol_on_helper(self):
        public = self.ds.folders.public
        sf1 = public.create_folder("level1")
        sf1a = public.create_folder("level1a")
        sf2 = sf1.create_folder("level2")

        top_level = self.ds.folders  # Testing over the helper class
        # Access by path []
        self.assertEqual(top_level[sf2.path].url, sf2.url)

        # Access by name
        self.assertEqual(top_level[sf1.name].url, sf1.url)

        # __iter__ method
        self.assertEqual([c.url for c in top_level],
            [c.url for c in public.children])

        # more dict methods
        self.assertEqual(set(dict(top_level).keys()), set(public.keys()))
        self.assertEqual({c.url for c in dict(top_level).values()},
            {c.url for c in public.children})
        # The right key corresponds to the correct value
        self.assertEqual(dict(top_level)[sf1.name].url, sf1.url)




# coding: utf-8

import pytest
import textwrap
from six import StringIO

from pycrunch.shoji import as_entity

from scrunch.mutable_dataset import get_mutable_dataset
from fixtures import BaseIntegrationTestCase


class TestBackFill(BaseIntegrationTestCase):
    def _prepare_ds(self, values):
        ds = self.site.datasets.create(
            as_entity({"name": "test_backfill_values"})).refresh()
        # We need a numeric PK
        pk = ds.variables.create(
            as_entity(
                {
                    "name": "pk",
                    "alias": "pk",
                    "type": "numeric",
                    "values": values["pk"],
                }
            )
        )

        # Create a categorical, note the segment of -1 in rows 4, 5, 6
        cat1 = ds.variables.create(
            as_entity(
                {
                    "name": "cat1",
                    "alias": "cat1",
                    "type": "categorical",
                    "categories": [
                        {
                            "id": 1,
                            "name": "cat 1",
                            "missing": False,
                            "numeric_value": None,
                        },
                        {
                            "id": 2,
                            "name": "cat 2",
                            "missing": False,
                            "numeric_value": None,
                        },
                        {
                            "id": 3,
                            "name": "cat 3",
                            "missing": False,
                            "numeric_value": None,
                        },
                        {
                            "id": -1,
                            "name": "No Data",
                            "missing": True,
                            "numeric_value": None,
                        },
                    ],
                    "values": values["cat1"],
                }
            )
        )

        # Create another categorical, note the segment of -1 in rows 4, 5, 6
        cat2 = ds.variables.create(
            as_entity(
                {
                    "name": "cat2",
                    "alias": "cat2",
                    "type": "categorical",
                    "categories": [
                        {
                            "id": 11,
                            "name": "cat 1",
                            "missing": False,
                            "numeric_value": None,
                        },
                        {
                            "id": 22,
                            "name": "cat 2",
                            "missing": False,
                            "numeric_value": None,
                        },
                        {
                            "id": 33,
                            "name": "cat 3",
                            "missing": False,
                            "numeric_value": None,
                        },
                        {
                            "id": -1,
                            "name": "No Data",
                            "missing": True,
                            "numeric_value": None,
                        },
                    ],
                    "values": values["cat2"],
                }
            )
        )
        cat3 = ds.variables.create(
            as_entity(
                {
                    "name": "cat3",
                    "alias": "cat3",
                    "type": "categorical",
                    "categories": [
                        {
                            "id": 1,
                            "name": "cat 1",
                            "missing": False,
                            "numeric_value": None,
                        },
                        {
                            "id": 2,
                            "name": "cat 2",
                            "missing": False,
                            "numeric_value": None,
                        },
                        {
                            "id": 3,
                            "name": "cat 3",
                            "missing": False,
                            "numeric_value": None,
                        },
                        {
                            "id": -1,
                            "name": "No Data",
                            "missing": True,
                            "numeric_value": None,
                        },
                    ],
                    "values": values["cat3"],
                }
            )
        )
        return ds

    def test_backfill_values(self):
        ds = self._prepare_ds({
            "pk": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
            "cat1": [1, 2, 3, -1, -1, -1, 1, 2, 3, 1],
            "cat2": [11, 22, 33, -1, -1, -1, 11, 22, 33, 11],
            "cat3": [1, 2, 3, -1, -1, -1, 1, 2, 3, 1],
        })
        csv_file = StringIO(textwrap.dedent("""pk,cat1,cat2
        4,1,22
        5,2,33
        6,3,11
        """))
        scrunch_dataset = get_mutable_dataset(ds.body.id, self.site)

        rows_expr = "pk >= 4 and pk <=6"
        scrunch_dataset.backfill_from_csv(["cat1", "cat2"], "pk", csv_file, rows_expr)

        data = ds.follow("table", "limit=10")["data"]
        vars = ds.variables.by("alias")
        assert data[vars["cat1"]["id"]] == [1, 2, 3, 1, 2, 3, 1, 2, 3, 1]
        assert data[vars["cat2"]["id"]] == [11, 22, 33, 22, 33, 11, 11, 22, 33, 11]

        assert len(data) == 4  # Only PK and cat1 and cat2 and cat3
        assert "folder" not in ds.folders.by("type")  # Nothing of type folder

        ds.delete()

    def test_backfill_on_subvars(self):
        ds = self._prepare_ds({
            "pk": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
            "cat1": [1, 2, 3, -1, -1, -1, 1, 2, 3, 1],
            "cat2": [11, 22, 33, -1, -1, -1, 11, 22, 33, 11],
            "cat3": [2, 3, 1, -1, -1, -1, 2, 3, 1, 2]
        })
        vars = ds.variables.by("alias")
        array = ds.variables.create(as_entity({
            "name": "array",
            "alias": "array",
            "type": "categorical_array",
            "subvariables": [vars["cat1"].entity_url, vars["cat3"].entity_url],
        })).refresh()

        csv_file = StringIO(textwrap.dedent("""pk,cat1,cat3
                4,1,2
                5,2,3
                6,3,1
                """))
        scrunch_dataset = get_mutable_dataset(ds.body.id, self.site)

        rows_expr = "pk >= 4 and pk <=6"
        scrunch_dataset.backfill_from_csv(["cat1", "cat3"], "pk", csv_file,
            rows_expr)

        data = ds.follow("table", "limit=10")["data"]
        assert data[array.body["id"]] == [
            [1, 2],
            [2, 3],
            [3, 1],
            [1, 2],
            [2, 3],
            [3, 1],
            [1, 2],
            [2, 3],
            [3, 1],
            [1, 2]
        ]

        ds.delete()

    def test_backfill_on_subvars_full_row(self):
        ds = self._prepare_ds({
            "pk": [1, 2, 3, 4, 5],
            "cat1": [1, 2, 3, -1, -1],
            "cat2": [11, 22, 33, -1, -1],
            "cat3": [2, 3, 1, -1, -1]
        })
        vars = ds.variables.by("alias")
        subvars = [vars["cat1"].entity_url, vars["cat2"].entity_url,  vars["cat3"].entity_url]
        array = ds.variables.create(as_entity({
            "name": "array",
            "alias": "array",
            "type": "categorical_array",
            "subvariables": subvars,
        })).refresh()

        csv_file = StringIO(textwrap.dedent("""pk,cat1,cat3
                1,1,2
                2,2,3
                3,3,1
                4,2,3
                5,2,1
                """))
        scrunch_dataset = get_mutable_dataset(ds.body.id, self.site)

        # Not including a row_filter, same as passing None
        scrunch_dataset.backfill_from_csv(["cat1", "cat3"], "pk", csv_file)

        assert ds.follow("table", "limit=5")["data"][array.body["id"]] == [
            [1, 1, 2],
            [2, 2, 3],
            [3, 3, 1],
            [2, {"?": -1}, 3],
            [2, {"?": -1}, 1]
        ]

        ds.delete()

    def test_backfill_on_non_missing(self):
        ds = self._prepare_ds({
            "pk": [1, 2, 3, 4, 5],
            "cat1": [1, 2, 3, 3, 3],
            "cat2": [11, 22, 33, 11, 22],
            "cat3": [1, 1, 1, 1, 1]
        })
        csv_file = StringIO(textwrap.dedent("""pk,cat1,cat3
                4,1,2
                5,2,3
                """))
        scrunch_dataset = get_mutable_dataset(ds.body.id, self.site)

        rows_expr = "pk >= 4 and pk <=5"
        scrunch_dataset.backfill_from_csv(["cat1", "cat3"], "pk", csv_file,
            rows_expr)

        vars = ds.variables.by("alias")
        data = ds.follow("table", "limit=10")["data"]
        assert data[vars["cat3"]["id"]] == [1, 1, 1, 2, 3]
        assert data[vars["cat1"]["id"]] == [1, 2, 3, 1, 2]

        ds.delete()

    def test_bad_csv(self):
        original_data = {
            "pk": [1, 2, 3, 4],
            "cat1": [1, -1, 2, 3],
            "cat2": [11, 22, 33, -1],
            "cat3": [1, -1, 3, -1],
        }
        ds = self._prepare_ds(original_data)
        csv_file = StringIO(textwrap.dedent("""pk,BOGUS,BAD
            2,1,22
        """))
        scrunch_dataset = get_mutable_dataset(ds.body.id, self.site)

        rows_expr = "pk == 2"
        with pytest.raises(ValueError) as err:
            scrunch_dataset.backfill_from_csv(["cat1", "cat2"], "pk", csv_file,
                rows_expr)
        assert err.value.args[0].startswith("Invalid data provided: Expected column ")

        # Verify that the backfill didn't proceed
        data = ds.follow("table", "limit=10")["data"]
        variables = ds.variables.by("alias")
        assert data[variables["cat1"]["id"]] == [1, {"?": -1}, 2, 3]
        assert data[variables["cat2"]["id"]] == [11, 22, 33, {"?": -1}]

        assert len(data) == 4  # Only PK and cat1 and cat2 and cat3
        assert "folder" not in ds.folders.by("type")  # Nothing of type folder

        ds.delete()

    def test_with_exclusion_filter(self):
        ds = self._prepare_ds({
            "pk": [1, 2, 3, 4, 5],
            "cat1": [1, 2, 3, 3, 3],
            "cat2": [11, 11, 11, 11, 11],
            "cat3": [1, 1, 1, 1, 1]
        })
        csv_file = StringIO(textwrap.dedent("""pk,cat1,cat3
                        4,1,2
                        5,2,3
                        """))
        scrunch_dataset = get_mutable_dataset(ds.body.id, self.site)

        excl = "pk == 4"
        scrunch_dataset.exclude(excl)
        rows_expr = "pk in [4, 5]"
        scrunch_dataset.backfill_from_csv(["cat1", "cat3"], "pk", csv_file,
            rows_expr)

        # Exclusion gets set after backfilling
        assert scrunch_dataset.get_exclusion() == excl

        ds.exclusion.patch({"expression": {}})
        varsiables = ds.variables.by("alias")
        data = ds.follow("table", "limit=10")["data"]
        assert data[varsiables["cat1"]["id"]] == [1, 2, 3, 1, 2]
        assert data[varsiables["cat3"]["id"]] == [1, 1, 1, 2, 3]

        ds.delete()

    def test_too_big_file(self):
        ds = self._prepare_ds({
            "pk": [1, 2, 3, 4, 5],
            "cat1": [1, 2, 3, 3, 3],
            "cat2": [11, 11, 11, 11, 11],
            "cat3": [1, 1, 1, 1, 1]
        })
        scrunch_dataset = get_mutable_dataset(ds.body.id, self.site)

        size_200MB = 200 * 2 ** 20
        csv_file = StringIO("x" * size_200MB)
        with pytest.raises(ValueError) as err:
            scrunch_dataset.backfill_from_csv(["cat1"], "pk", csv_file, None)
        assert err.value.args[0] == "Max CSV allowed size is currently 150MB"



# coding: utf-8

import pycrunch
from scrunch.helpers import shoji_view_wrapper
from scrunch.scripts import ScriptExecutionError, SystemScript
from scrunch.connections import _default_connection
from scrunch.datasets import Project


class AccountProjects:
    """
    Handles the list of top level projects for the account. Note that this can
    403 if the authenticated user does not have permissions to access the
    projects.
    """

    def __init__(self, resource):
        self.resource = resource

    def by_name(self, name):
        project_res = self.resource.by("name")[name].entity
        return Project(project_res)

    def by_id(self, _id):
        project_res = self.resource.by("id")[_id].entity
        return Project(project_res)


class Account:
    """
    Represents the account for the authenticated user.

    If follows `/api/account/`
    """

    def __init__(self, resource):
        self.resource = resource

    @property
    def name(self):
        return self.resource["body"]["name"]

    @property
    def id(self):
        return self.resource["body"]["id"]

    @classmethod
    def current_account(cls, connection=None):
        site_root = _default_connection(connection)
        act_res = site_root.account
        return cls(act_res)

    def execute(self, script_body, strict_subvariable_syntax=None):
        """Will run a system script on this account."""
        # The account execution endpoint is a shoji:view
        system_script = SystemScript(self.resource)
        return system_script.execute(script_body, strict_subvariable_syntax)

    @property
    def projects(self):
        act_projects = self.resource.projects
        return AccountProjects(act_projects)




# coding: utf-8
import os


import pytest
from pycrunch.importing import Importer
from pycrunch.shoji import as_entity

from fixtures import BaseIntegrationTestCase, MR_CATS
from scrunch.mutable_dataset import get_mutable_dataset


@pytest.mark.skipif(os.environ.get("LOCAL_INTEGRATION") is None, reason="Do not run this test during CI/CD")
class TestExpressions(BaseIntegrationTestCase):

    def _create_mr_dataset(self, name, rows):
        _dataset_metadata = {
            "mr_variable": {
                "name": "Multiple Response",
                "type": "multiple_response",
                "categories": MR_CATS,
                "subreferences": [{
                    "name": "Response 1",
                    "alias": "response_1"
                }, {
                    "name": "Response 2",
                    "alias": "response_2"
                }, {
                    "name": "Response 3",
                    "alias": "response_3"
                }]
            },
        }
        ds = self.site.datasets.create({
            'element': 'shoji:entity',
            'body': {
                'name': name,
                'table': {
                    'element': 'crunch:table',
                    'metadata': _dataset_metadata
                },
            }
        }).refresh()
        Importer().append_rows(ds, rows)
        scrunch_dataset = get_mutable_dataset(ds.body.id, self.site)
        return ds, scrunch_dataset

    def test_multiple_response_all_add_filter_value(self):
        ds_rows = [
            ["response_1", "response_2", "response_3"],
            [2, 2, 1],
            [1, 2, 2],
            [1, 1, 1]
        ]
        ds, scrunch_dataset = self._create_mr_dataset('test_mr_any', ds_rows)
        _filter = "mr_variable.all([1])"
        try:
            resp = scrunch_dataset.add_filter(name='filter_1', expr=_filter)
            data = ds.follow("table", "limit=20&filter={}".format(resp.resource.self))['data']
            ds_variables = ds.variables.by("alias")
            mr_variable_id = ds_variables["mr_variable"].id
            assert data[mr_variable_id] == [
                [1, 1, 1]
            ]
        finally:
            # cleanup
            ds.delete()

    def test_multiple_response_any_add_filter_single_subvar(self):
        ds_rows = [
            ["response_1", "response_2", "response_3"],
            [2, 2, 1],
            [1, 2, 2],
            [2, 1, 1]
        ]
        ds, scrunch_dataset = self._create_mr_dataset('test_mr_any', ds_rows)
        _filter = "mr_variable.any([response_1])"
        try:
            resp = scrunch_dataset.add_filter(name='filter_1', expr=_filter)
            data = ds.follow("table", "limit=20&filter={}".format(resp.resource.self))['data']
            ds_variables = ds.variables.by("alias")
            mr_variable_id = ds_variables["mr_variable"].id
            assert data[mr_variable_id] == [
                [1, 2, 2]
            ]
        finally:
            # cleanup
            ds.delete()

    def test_multiple_response_any_add_filter_subvar(self):
        ds_rows = [
            ["response_1", "response_2", "response_3"],
            [2, 2, 1],
            [2, 2, 2],
            [2, 1, 1]
        ]
        ds, scrunch_dataset = self._create_mr_dataset('test_mr_any', ds_rows)
        _filter = "mr_variable.any([response_1, response_2])"
        try:
            resp = scrunch_dataset.add_filter(name='filter_1', expr=_filter)
            data = ds.follow("table", "limit=20&filter={}".format(resp.resource.self))['data']
            ds_variables = ds.variables.by("alias")
            mr_variable_id = ds_variables["mr_variable"].id
            assert data[mr_variable_id] == [
                [2, 1, 1]
             ]
        finally:
            # cleanup
            ds.delete()

    def test_categorical_array_any_add_filter(self):
        ds = self.site.datasets.create(as_entity({"name": "test_any_categorical_add_filter"})).refresh()
        ds.variables.create(as_entity({
            "name": "Categorical Var",
            "alias": "categorical_var",
            "type": "categorical_array",
            "categories": [
                {
                    'id': 1,
                    'missing': False,
                    'name': 'Very interested',
                    'numeric_value': 1
                },
                {
                    'id': 2,
                    'missing': False,
                    'name': 'Somewhat interested',
                    'numeric_value': 2
                },
                {
                    'id': 3,
                    'missing': False,
                    'name': 'A little interested',
                    'numeric_value': 3
                },
                {
                    'id': -1,
                    'missing': True,
                    'name': 'No Data',
                    'numeric_value': None
                }
            ],
            'subvariables': [
                {
                    'alias': 'response_1',
                    'name': 'Response1'
                },
                {
                    'alias': 'response_2',
                    'name': 'Response2'
                },
                {
                    'alias': 'response_3',
                    'name': 'Response3'
                },
            ],
            "values": [
                [1, 3, -1],
                [2, 1, 1],
                [2, 3, 2]
            ]
        }))
        scrunch_dataset = get_mutable_dataset(ds.body.id, self.site)
        _filter = "categorical_var.any([1])"
        try:
            resp = scrunch_dataset.add_filter(name='filter_1', expr=_filter)
            data = ds.follow("table", "limit=20&filter={}".format(resp.resource.self))['data']
            ds_variables = ds.variables.by("alias")
            cat_var_variable_id = ds_variables["categorical_var"].id
            assert data[cat_var_variable_id] == [
                [1, 3, {"?": -1}],
                [2, 1, 1]
            ]
            assert resp.resource.body.entity.body.definition == {
                "function": "or",
                "args": [
                    {
                        "function": "in",
                        "args": [
                            {
                                "var": "categorical_var",
                                "axes": ["response_1"]
                            },
                            {
                                "value": [1]
                            }
                        ]
                    },
                    {
                        "function": "in",
                        "args": [{
                            "var": "categorical_var",
                            "axes": ["response_2"]
                        },
                            {"value": [1]
                             }
                        ]
                    },
                    {
                        "function": "in",
                        "args": [{
                            "var": "categorical_var",
                            "axes": ["response_3"]
                        },
                            {"value": [1]
                             }
                        ]
                    }
                ]
            }
        finally:
            # cleanup
            ds.delete()

    def test_categorical_array_any_w_bracket_subvar(self):
        ds = self.site.datasets.create(as_entity({"name": "test_any_categorical_w_bracket_add_filter"})).refresh()
        cat_var = ds.variables.create(as_entity({
            "name": "Categorical Var",
            "alias": "categorical_var",
            "type": "categorical_array",
            "categories": [
                {
                    'id': 1,
                    'missing': False,
                    'name': 'Very interested',
                    'numeric_value': 1
                },
                {
                    'id': 2,
                    'missing': False,
                    'name': 'Somewhat interested',
                    'numeric_value': 2
                },
                {
                    'id': 3,
                    'missing': False,
                    'name': 'A little interested',
                    'numeric_value': 3
                },
                {
                    'id': -1,
                    'missing': True,
                    'name': 'No Data',
                    'numeric_value': None
                }
            ],
            'subvariables': [
                {
                    'alias': 'response_1',
                    'name': 'Response1'
                },
                {
                    'alias': 'response_2',
                    'name': 'Response2'
                },
                {
                    'alias': 'response_3',
                    'name': 'Response3'
                },
            ],
            "values": [
                [1, 3, -1],
                [2, 1, 1],
                [2, 3, 2]
            ]
        }))
        scrunch_dataset = get_mutable_dataset(ds.body.id, self.site)
        _filter = "categorical_var[response_1].any([1])"
        try:
            resp = scrunch_dataset.add_filter(name='filter_1', expr=_filter)
            data = ds.follow("table", "limit=20&filter={}".format(resp.resource.self))['data']
            ds_variables = ds.variables.by("alias")
            cat_var_variable_id = ds_variables["categorical_var"].id
            assert data[cat_var_variable_id] == [
                [1, 3, {"?": -1}]
            ]
            assert resp.resource.body.entity.body.definition == {
                "function": "in",
                "args": [
                    {
                        "var": "categorical_var",
                        "axes": ["response_1"]
                    },
                    {
                        "value": [1]
                    }
                ]
            }
        finally:
            # cleanup
            ds.delete()

    def test_append_dataset_any_filter_multiple_response(self):
        ds_rows = [
            ["response_1", "response_2", "response_3"],
            [1, 2, 1],
            [1, 2, 2],
            [1, 1, 1]
        ]
        ds_to_append_rows = [
            ["response_1", "response_2", "response_3"],
            [2, 1, 2],
            [2, 2, 1],
            [2, 2, 1]
        ]
        ds, scrunch_dataset = self._create_mr_dataset('test_mr_any_subvar', ds_rows)
        ds_to_append, scrunch_dataset_to_append = self._create_mr_dataset(
            'test_mr_any_to_append_subvar',
            ds_to_append_rows
        )
        # This filter should get only the rows that have the mr_response variable with the value 1
        # at the same time for both response_1 and response_2
        _filter = "mr_variable.any([response_1, response_2])"
        try:
            scrunch_dataset.append_dataset(scrunch_dataset_to_append, filter=_filter)
            ds_variables = ds.variables.by("alias")
            mr_variable_id = ds_variables["mr_variable"].id
            data = ds.follow("table", "limit=20")['data']
            assert data[mr_variable_id] == [
                [1, 2, 1],
                [1, 2, 2],
                [1, 1, 1],
                [2, 1, 2],
            ]
        finally:
            # cleanup
            ds.delete()
            ds_to_append.delete()

    def test_append_dataset_any_filter_multiple_response_single_subvar(self):
        ds_rows = [
            ["response_1", "response_2", "response_3"],
            [1, 2, 1],
            [1, 2, 2],
            [1, 1, 1]
        ]
        ds_to_append_rows = [
            ["response_1", "response_2", "response_3"],
            [1, 1, 2],
            [2, 2, 1],
            [1, 1, 1]
        ]
        ds, scrunch_dataset = self._create_mr_dataset('test_mr_any_subvar', ds_rows)
        ds_to_append, scrunch_dataset_to_append = self._create_mr_dataset(
            'test_mr_any_to_append_subvar',
            ds_to_append_rows
        )
        # This filter should get only the rows that have the mr_response variable with the value 1 (selected)
        # for response_1 (not the 2nd row in this test)
        _filter = "mr_variable.any([response_1])"
        try:
            scrunch_dataset.append_dataset(scrunch_dataset_to_append, filter=_filter)
            ds_variables = ds.variables.by("alias")
            mr_variable_id = ds_variables["mr_variable"].id
            data = ds.follow("table", "limit=20")['data']
            assert data[mr_variable_id] == [
                [1, 2, 1],
                [1, 2, 2],
                [1, 1, 1],
                [1, 1, 2],
                [1, 1, 1]
            ]
        finally:
            # cleanup
            ds.delete()
            ds_to_append.delete()

    def test_categorical_any_add_filter_value(self):
        ds = self.site.datasets.create(as_entity({"name": "test_any_categorical_filter"})).refresh()
        categories = [
            {"id": 1, "name": "One", "missing": False, "numeric_value": None},
            {"id": 2, "name": "Two", "missing": False, "numeric_value": None},
            {"id": 3, "name": "Three", "missing": False, "numeric_value": None},
            {"id": -1, "name": "No Data", "missing": True, "numeric_value": None}
        ]

        ds.variables.create(as_entity({
            "name": "my_cat",
            "alias": "my_cat",
            "type": "categorical",
            "categories": categories,
            "values": [1, -1, 2, 3, 3, 1]
        }))

        scrunch_dataset = get_mutable_dataset(ds.body.id, self.site)
        _filter = "my_cat.any([1])"
        try:
            scrunch_dataset.add_filter(name='filter_1', expr=_filter)
            # Adding the filter as exclusion. In this case, we are expecting the opposite
            # of the filter since it is an exclusion one
            scrunch_dataset.exclude(_filter)
            data = ds.follow("table", "limit=20")['data']
            ds_variables = ds.variables.by("alias")
            variable_id = ds_variables["my_cat"].id
            assert data[variable_id] == [{'?': -1}, 2, 3, 3]
        finally:
            # cleanup
            ds.delete()

    def test_categorical_any_add_filter_multiple_values(self):
        ds = self.site.datasets.create(as_entity({"name": "test_any_categorical_filter_multiple_values"})).refresh()
        categories = [
            {"id": 1, "name": "One", "missing": False, "numeric_value": None},
            {"id": 2, "name": "Two", "missing": False, "numeric_value": None},
            {"id": 3, "name": "Three", "missing": False, "numeric_value": None},
            {"id": -1, "name": "No Data", "missing": True, "numeric_value": None}
        ]

        ds.variables.create(as_entity({
            "name": "my_cat",
            "alias": "my_cat",
            "type": "categorical",
            "categories": categories,
            "values": [1, -1, 2, 3, 3, 1]
        }))

        scrunch_dataset = get_mutable_dataset(ds.body.id, self.site)
        _filter = "my_cat.any([1, 3])"
        try:
            resp = scrunch_dataset.add_filter(name='filter_1', expr=_filter)
            data = ds.follow("table", "limit=20&filter={}".format(resp.resource.self))['data']
            ds_variables = ds.variables.by("alias")
            cat_var_variable_id = ds_variables["my_cat"].id
            assert data[cat_var_variable_id] == [1, 3, 3, 1]
        finally:
            # cleanup
            ds.delete()



from pycrunch.version import __version__ as pycrunch_version
from pycrunch import connect as _connect
from pycrunch.elements import ElementSession

from .version import __version__


class ScrunchSession(ElementSession):
    headers = {
        "user-agent": "scrunch/%s (pycrunch/%s)" % (__version__, pycrunch_version)
    }


class ScrunchSSLUnsafeSession(ScrunchSession):
    """
    A subclass of `ScrunchSession` that skips SSL certificate validation
    when trying to connect to the API server. Useful for local testing.
    """
    def __init__(self, *args, **kwargs):
        super(ScrunchSSLUnsafeSession, self).__init__(*args, **kwargs)
        self.verify = False


FLAGS_TO_CHECK = {"old_projects_order", "clients_strict_subvariable_syntax"}


def set_feature_flags(site):
    feature_flags = {
        flag_name: site.follow('feature_flag',
                               'feature_name=%s' % flag_name).value['active']
        for flag_name in FLAGS_TO_CHECK
    }
    setattr(site.session, 'feature_flags', feature_flags)
    return site


def connect(*args, **kwargs):
    _site = _connect(session_class=ScrunchSession, *args, **kwargs)
    _site = set_feature_flags(_site)
    return _site



import six

from pycrunch.cubes import fetch_cube, count
from scrunch.datasets import Variable
from scrunch.expressions import parse_expr, process_expr

from cr.cube.cube import Cube


def variable_to_url(variable, dataset):
    """
    Get a variable alias, instance or url and return
    the url
    """
    if isinstance(variable, Variable):
        return variable.url
    elif isinstance(variable, six.string_types):
        if variable in dataset.keys():
            return dataset[variable].url
    return variable


def crtabs(dataset, variables, weight=None, filter_=None, transforms=None, **measures):
    """Return Cube representation of crosstab.

    :param dataset: Dataset instance
    :param variables: List of variable urls, aliases or Variable instances
    :param weight: Scrunch variable instance, alias or url
    :param filter_: Scrunch filter expression
    :param transforms: cr.cube transforms dictionary
    """
    variables = [variable_to_url(var, dataset) for var in variables]
    if weight is not None:
        weight = variable_to_url(weight, dataset)
    if filter_ is not None:
        filter_ = process_expr(parse_expr(filter_), dataset.resource)

    # if six.PY2:
    #     raise DeprecationWarning("Crunch Cube isn't supported in Python 2.x")

    return Cube(
        fetch_cube(
            dataset.resource,
            variables,
            count=count(),
            weight=weight,
            filter=filter_,
            **measures
        ),
        transforms=transforms,
    )



# coding: utf-8
import json

import six.moves.urllib as urllib
from six import PY2

from pycrunch.lemonpy import URL
from pycrunch.progress import DefaultProgressTracking
from pycrunch.shoji import wait_progress
from scrunch.helpers import download_file


class SubEntity:
    """
    A pycrunch.shoji.Entity directly related to a Dataset.
    For example; filters, decks
    """
    _MUTABLE_ATTRIBUTES = set()
    _IMMUTABLE_ATTRIBUTES = set()
    _ENTITY_ATTRIBUTES = set()

    def __init__(self, shoji_tuple):
        self.resource = shoji_tuple.entity

    def __getattr__(self, item):
        if item in self._ENTITY_ATTRIBUTES:
            return self.resource.body[item]
        raise AttributeError(
            '{} has no attribute {}'.format(self.__class__.__name__, item))

    def __repr__(self):
        if PY2:
            name = self.name.encode("ascii", "replace")
        else:
            name = self.name
        return u"<{}: name='{}'; id='{}'>".format(
            self.__class__.__name__, name, self.id)

    def __str__(self):
        return self.__repr__()

    def edit(self, **kwargs):
        for key in kwargs:
            if key not in self._MUTABLE_ATTRIBUTES:
                raise AttributeError(
                    "Can't edit attribute {} of {}".format(key, self.name))
        return self.resource.edit(**kwargs)

    def remove(self):
        self.resource.delete()

    def delete(self):
        """
        Aalias to remove method
        """
        self.remove()


class Filter(SubEntity):
    """
    A pycrunch.shoji.Entity for Dataset filters
    """
    _MUTABLE_ATTRIBUTES = {'name', 'template', 'is_public', 'owner_id'}
    _IMMUTABLE_ATTRIBUTES = {'id', }
    _ENTITY_ATTRIBUTES = _MUTABLE_ATTRIBUTES | _IMMUTABLE_ATTRIBUTES


class Multitable(SubEntity):
    """
    A pycrunch.shoji.Entity for Multitables
    """
    _MUTABLE_ATTRIBUTES = {'name', 'template', 'is_public'}
    _IMMUTABLE_ATTRIBUTES = {'id', }
    _ENTITY_ATTRIBUTES = _MUTABLE_ATTRIBUTES | _IMMUTABLE_ATTRIBUTES

    def __init__(self, shoji_tuple, ds):
        self.resource = shoji_tuple.entity
        # a dataset instance to make things simpler
        self.ds = ds

    def query_cube(self):
        """
        http://docs.crunch.io/#more-complex-multitable-templates
        """
        raise NotImplementedError

    def export_tabbook(self, format, progress_tracker=None, filter=None,
                       where=None, options=None, weight=False):
        """
        An adaption of https://github.com/Crunch-io/pycrunch/blob/master/pycrunch/exporting.py
        to Multitables exports (tabbboks)
        """
        payload = {}

        # add filter to multitable
        if filter:
            if isinstance(filter, Filter):
                payload['filter'] = [{'filter': filter.resource.self}]
            else:
                raise ValueError('filter param must be a Filter instance')

        if options and isinstance(options, dict):
            payload['options'] = options

        if where:
            if isinstance(where, list):
                id_vars = []
                for var in where:
                    id_vars.append(self.ds[var].url)
                # Now build the payload with selected variables
                payload['where'] = {
                    'function': 'make_frame',
                    'args': [{
                        'map': {
                            x: {'variable': x} for x in id_vars
                        }
                    }]
                }
            else:
                raise ValueError('where param must be a list of variable names')

        if weight:
            payload['weight'] = self.ds[weight].url
        if weight is None:
            payload['weight'] = None

        session = self.resource.session
        endpoint = self.resource.views['tabbook']

        # in case of json format, we need to return the json response
        if format == 'json':
            r = session.post(
                endpoint,
                json.dumps(payload),
                headers={'Accept': 'application/json'})
        else:
            r = session.post(endpoint, json.dumps(payload))
        dest_file = URL(r.headers['Location'], '')
        if r.status_code == 202:
            try:
                r.payload['value']
            except Exception:
                # Not a progress API just return the incomplete entity.
                # User will refresh it.
                pass
            else:
                # We have a progress_url, wait for completion
                wait_progress(r, session, progress_tracker)
        return dest_file

    def export(self, path, format='xlsx', timeout=None, filter=None,
               where=None, options=None, **kwargs):
        """
        A tabbook export: http://docs.crunch.io/#tab-books
        Exports data as csv to the given path or as a JSON response
        :params: http://docs.crunch.io/#post-body-parameters
        :path: Local Filesystem path to save the file to
        :filter: Name of Filter instance of a Dataset's filter
        :where: list of variables to include; ['varA', 'varB']
        :options: Display options as python dictionary
        :weight: Name of the weight_variable
        """
        if format not in ['xlsx', 'json']:
            raise ValueError("Format can only be 'json' or 'xlxs'")
        progress_tracker = DefaultProgressTracking(timeout)
        tabbook_args = dict(
            format=format,
            progress_tracker=progress_tracker,
            filter=filter,
            where=where,
            options=options,
        )
        if 'weight' in kwargs:
            tabbook_args['weight'] = kwargs['weight']
        else:
            tabbook_args['weight'] = False
        url = self.export_tabbook(**tabbook_args)
        download_file(url, path)


class Deck(SubEntity):
    """
    A pycrunch.shoji.Entity for Dataset decks
    """
    _MUTABLE_ATTRIBUTES = {'name', 'description', 'is_public',
                           'owner_id', 'owner_name'}
    _IMMUTABLE_ATTRIBUTES = {'id', 'creation_time', 'slides'}
    _ENTITY_ATTRIBUTES = _MUTABLE_ATTRIBUTES | _IMMUTABLE_ATTRIBUTES

    def __repr__(self):
        return "<{}: id='{}'; name='{}'>".format(
            self.__class__.__name__, self.id, self.name)

    @property
    def slides(self):
        _slides = {}
        for slide in self.resource.slides.index.values():
            slide_inst = Slide(slide)
            _slides[slide_inst.id] = slide_inst
        return _slides

    @slides.setter
    def slides(self, _):
        # Protect `slides` property from direct modifications.
        raise TypeError('Use add_decks method to add a new deck')

    def xlsx_export(self):
        raise NotImplementedError


class Slide(SubEntity):
    """
    A pycrunch.shoji.Entity for a Slide
    """
    _MUTABLE_ATTRIBUTES = {'display_settings', 'analysis_url', 'title',
                           'subtitle', 'deck_id'}
    _IMMUTABLE_ATTRIBUTES = {'id', 'dataset_id', 'version', 'analyses'}
    _ENTITY_ATTRIBUTES = _MUTABLE_ATTRIBUTES | _IMMUTABLE_ATTRIBUTES

    def __repr__(self):
        return "<{}: id='{}'; title='{}'>".format(
            self.__class__.__name__, self.id, self.title)

    @property
    def analyses(self):
        _analyses = {}
        for url, a in self.resource.analyses.index.items():
            id = url.split('/')[-2]
            _analyses[id] = Analysis(a, id)
        return _analyses

    @analyses.setter
    def analyses(self, _):
        # Protect `analyses` property from direct modifications.
        raise TypeError('Use add_decks method to add a new deck')


class Analysis:
    """
    A cube query JSON representation and transcriber
    """

    def __init__(self, shoji_tuple, id):
        self.resource = shoji_tuple.entity
        self.id = id

    def __getitem__(self, item):
        try:
            return self.resource.body[item]
        except AttributeError:
            raise AttributeError(
                '{} has no attribute {}'.format(
                    self.__class__.__name__, item))

    def __repr__(self):
        return "<{}: id='{}'>".format(self.__class__.__name__, self.id)

    def query_cube(self, ds):
        """
        :param: ds: Dataset() instance
        out of the current instance GET a cube query
        :return: a shoji:view json instance
        """
        json_string = self['query'].json
        # this process removes newlines
        dict_obj = json.loads(json_string)
        resp = ds.resource.follow(
            'cube',
            urllib.parse.urlencode({'query': json.dumps(dict_obj)})
        )
        return resp.json

    def as_dataframe(self):
        """
        It would be awesome to visualize a table or Analysis
        as a pandas.Dataframe or even Numpy Array
        """
        raise NotImplementedError
